package de.hsel.itech.db;

import de.hsel.itech.db.pojo.Address;
import de.hsel.itech.db.pojo.User;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * handles all database stuff for addresses of users.
 *
 * @author Johannes Theiner
 * @version 0.1
 * @since 0.5
 **/
public class AddressDB {

    private Database database;

    AddressDB(Database database) {
        this.database = database;
    }

    /**
     * getAll all addresses of specified user.
     *
     * @param user user to getAll addresses from
     * @return list of users addresses
     */
    @Nullable
    public List<Address> getAll(User user) {
        if(user == null) return null;
        Map.Entry<ResultSet, Connection> entry = database.getResultSetsByValue(database.tableUserAddress, "user", String.valueOf(user.getId()));
        if(entry == null) return null;

        List<Address> list = new ArrayList<>();

        try {
            ResultSet rs = entry.getKey();
            while(rs.next()) {
                list.add(get(rs.getLong("address"), user));
            }

            entry.getValue().close();
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return list;
    }

    /**
     * getAll specific address.
     *
     * @param id address id
     * @return specific address
     */
    @Nullable
    public Address get(long id, User user) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tableAddress, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            rs.beforeFirst();
            Address address = null;
            while(rs.next()) {
                address = new Address(user, id, rs.getString("name"), rs.getString("street"), rs.getString("number"), rs.getInt("zipCode"), rs.getString("city"));
            }
            entry.getValue().close();

            return address;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }


    /**
     * insert address into database.
     *
     * @param address address to insert
     * @return insert count
     */
    public int insert(@NotNull Address address) {
        Connection connection = database.getConnection();
        int insertCount = 0;

        try {
            assert connection != null;
            PreparedStatement statement = connection.prepareStatement("INSERT INTO address (name, street, number, zipCode, city) VALUES (?, ?, ?, ?, ?)", Statement.RETURN_GENERATED_KEYS);
            statement.setString(1, address.getName());
            statement.setString(2, address.getStreet());
            statement.setString(3, address.getNumber());
            statement.setInt(4, address.getZipCode());
            statement.setString(5, address.getCity());

            insertCount += statement.executeUpdate();

            ResultSet keys = statement.getGeneratedKeys();
            Long id = null;
            while(keys.next()) {
                id = keys.getLong("id");
            }
            if(id != null) {
                PreparedStatement relationStatement = connection.prepareStatement("INSERT INTO user_address (user, address) VALUES (?, ?)");
                relationStatement.setLong(1, address.getUser().getId());
                relationStatement.setLong(2, id);

                insertCount += relationStatement.executeUpdate();
            }

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return insertCount;
    }

    /**
     * delete address from database.
     *
     * @param address address to delete
     * @return deletion count
     */
    public int delete(@NotNull Address address) {
        return database.delete(address.getId(), database.tableUserAddress);
    }
}