package de.hsel.itech.db;

import de.hsel.itech.db.pojo.Author;
import de.hsel.itech.db.pojo.Book;
import de.hsel.itech.db.pojo.Category;
import de.hsel.itech.db.pojo.Publisher;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.*;
import java.time.Year;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author Johannes Theiner
 * @version 0.1
 * @since 0.5
 **/
public class BookDB {
    private Database database;

    BookDB(Database database) {
        this.database = database;
    }

    /**
     * gets tableBook by id.
     *
     * @param id ID from database
     * @return {@link de.hsel.itech.db.pojo.Book}
     */
    @Nullable
    public Book get(long id) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tableBook, id);
        assert entry != null;
        try {
            entry.getKey().beforeFirst();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return build(entry);
    }

    /**
     * builds tableBook from entry.
     *
     * @param entry Map.Entry
     * @return BookDB
     */
    @Nullable
    private Book build(@NotNull Map.Entry<ResultSet, Connection> entry) {

        Book book = null;
        ResultSet rs = entry.getKey();
        try {
            rs.beforeFirst();
            while(rs.next()) {
                Category category = database.category().get(rs.getLong(database.tableCategory));
                Publisher publisher = database.publisher().get(rs.getLong(database.tablePublisher));
                List<Integer> authorIds = database.author().getAll(rs.getLong("id"));
                assert category != null;
                assert publisher != null;
                assert authorIds != null;

                List<Author> authors = new ArrayList<>();
                for (int i : authorIds) {
                    Author author = database.author().get(i);
                    assert author != null;
                    authors.add(author);
                }
                book = new Book(rs.getLong("id"), rs.getLong("isbn"), authors, publisher, category, rs.getString("title"),
                        Year.of(rs.getInt("year")), rs.getInt("price"), rs.getString("description"),
                        rs.getString("image"));
            }

            entry.getValue().close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return book;
    }

    /**
     * gets all books.
     *
     * @return {@link java.util.List}
     */
    @Nullable
    public List<Book> getAll() {
        List<Long> ids = database.getIds(database.tableBook);
        List<Book> books = new ArrayList<>();
        for (long id : ids) {
            books.add(get(id));
        }
        return books;
    }

    /**
     * gets books by tableCategory.
     *
     * @param category tableCategory id
     * @return {@link java.util.List}
     */
    @Nullable
    public List<Book> getAll(long category) {
        Connection connection = database.getConnection();
        assert connection != null;

        List<Book> books = new ArrayList<>();
        try {
            PreparedStatement statement = connection.prepareStatement("SELECT (id) FROM book WHERE category = ?");
            statement.setLong(1, category);

            ResultSet resultSet = statement.executeQuery();

            while (resultSet.next()) {
                books.add(get(resultSet.getLong("id")));
            }

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return books;
    }

    /**
     * getAll books by tableCategory.
     *
     * @param category Category object
     * @return {@link java.util.List}
     */
    @Nullable
    public List<Book> getAll(@NotNull Category category) {
        return getAll(category.getId());
    }

    /**
     * exists tableBook with isbn ?
     *
     * @param isbn isbn
     * @return true if tableBook with isbn exists
     */
    public boolean exists(long isbn) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetByValue(database.tableBook, "isbn", isbn);
        try {
            if (entry != null && entry.getKey() != null) {
                entry.getValue().close();
                return true;
            }
        } catch (SQLException ex) {
            ex.printStackTrace();
        }
        return false;
    }

    /**
     * deletes tableBook from database.
     *
     * @param id tableBook id
     * @return deletion count
     */
    public int delete(long id) {
        int deleteCount = 0;

        try {
            Connection connection = database.getConnection();
            assert connection != null;

            PreparedStatement statement = connection.prepareStatement("DELETE FROM author_book WHERE book = ?");
            statement.setLong(1, id);
            deleteCount = +statement.executeUpdate();

            connection.close();
        } catch (SQLException ex) {
            ex.printStackTrace();
        }

        deleteCount += database.delete(id, database.tableBook);

        return deleteCount;
    }


    /**
     * Inserts tableBook into database.
     * authors, publishers, and tableCategory have to exist.
     *
     * @param book BookDB
     * @return insert count
     */
    public int insert(@NotNull Book book) {
        Connection connection = database.getConnection();
        assert connection != null;

        int insertCount = 0;

        //tableAuthor exists ?
        List<Author> authors = new ArrayList<>();
        for (Author author : book.getAuthors()) {
            if (author.getId() == 0) {
                Author dbAuthor = database.author().get(author.getName());
                if (dbAuthor == null) {
                    insertCount += database.author().insert(author);
                    dbAuthor = database.author().get(author.getName());
                }
                authors.add(dbAuthor);
            } else authors.add(author);
        }

        //tablePublisher exists ?
        Publisher publisher;
        if (book.getPublisher().getId() == 0) {
            Publisher dbPublisher = database.publisher().get(book.getPublisher().getName());
            if (dbPublisher != null)
                publisher = dbPublisher;
            else {
                insertCount += database.publisher().insert(new Publisher(book.getPublisher().getName()));
                publisher = database.publisher().get(book.getPublisher().getName());
            }
        } else publisher = book.getPublisher();

        //tableCategory exists ?
        Category category;
        if (book.getCategory().getId() == 0) {
            Category dbCategory = database.category().get(book.getCategory().getName());
            if (dbCategory != null)
                category = dbCategory;
            else {
                insertCount += database.category().insert(new Category(book.getCategory().getName(), database.getRandomImage()));
                category = database.category().get(book.getCategory().getName());
            }
        } else category = book.getCategory();

        assert publisher != null;
        assert category != null;

        //insert tableBook
        try {
            PreparedStatement statement = connection.prepareStatement("INSERT INTO book(isbn, title, description, " +
                    "price, year, publisher, category, image) VALUES (?, ?, ?, ?, ?, ?, ?, ?)", Statement.RETURN_GENERATED_KEYS);
            statement.setLong(1, book.getIsbn());
            statement.setString(2, book.getTitle());
            statement.setString(3, book.getDescription());
            statement.setInt(4, book.getPrice());
            statement.setInt(5, book.getYear().getValue());
            statement.setLong(6, publisher.getId());
            statement.setLong(7, category.getId());
            statement.setString(8, book.getImage());
            insertCount += statement.executeUpdate();
            ResultSet resultSet = statement.getGeneratedKeys();
            long lastId = -1;
            while (resultSet.next()) {
                lastId = resultSet.getLong("id");
            }

            //add tableBook to tableAuthor
            for (Author author : authors) {
                PreparedStatement authorStatement = connection.prepareStatement("INSERT INTO " + database.tableAuthorBook + " (author, book) VALUES (?, ?);");
                authorStatement.setLong(1, author.getId());
                authorStatement.setLong(2, lastId);
                insertCount += authorStatement.executeUpdate();
            }

            connection.close();
        } catch (SQLException ex) {
            ex.printStackTrace();
        }
        return insertCount;
    }
}