package de.hsel.itech.db;


import de.hsel.itech.db.pojo.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.*;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Johannes Theiner
 * @version 0.1
 * @since 0.5
 */
public class PaymentDB {

    private Database database;

    PaymentDB(Database database) {
        this.database = database;
    }

    @Nullable
    private PaymentType getType(long id) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tablePaymentType, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            PaymentType type = null;
            rs.beforeFirst();
            while(rs.next()) {
                type = new PaymentType(id, rs.getString("name"));
            }
            entry.getValue().close();

            return type;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    @NotNull
    public List<? extends Payment> getAll(@NotNull User user) {
        List<Payment> list = new ArrayList<>();
        Map.Entry<ResultSet, Connection> entry = database.getResultSetByValue(database.tableUserPayment, "user", user.getId());
        if(entry == null) return list;

        try {
            ResultSet rs = entry.getKey();
            rs.beforeFirst();
            while(rs.next()) {
                list.add(get(rs.getLong("id"), user));
            }

            entry.getValue().close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return list;
    }

    @Nullable
    public Payment get(long id, @NotNull User user) {
        Map.Entry<User, PaymentType> metadata = getMetadata(id);
        assert metadata != null;

        switch (metadata.getValue().getName()) {
            case "PayPal":
                return getPayPal(id, user);
            case "CreditCard":
                return getCreditCard(id, user);
            case "DebitCard":
                return getDebitCard(id, user);
            case "Invoice":
                return getInvoice(id, user);
        }
        return null;
    }

    @Nullable
    private PayPalPayment getPayPal(long id, @NotNull User user) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tablePaymentPayPal, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            PayPalPayment payment = null;
            rs.beforeFirst();
            while(rs.next()) {
                payment = new PayPalPayment(id, user, rs.getString("mail"), rs.getString("auth"));
            }
            entry.getValue().close();

            return payment;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Nullable
    private CreditCardPayment getCreditCard(long id, @NotNull User user) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tablePaymentCredit, id);
        assert entry != null;


        try {
            ResultSet rs = entry.getKey();
            CreditCardPayment payment = null;
            rs.beforeFirst();
            while(rs.next()) {
                payment = new CreditCardPayment(id, user, rs.getInt("number"), rs.getString("owner"), rs.getDate("expiration"), rs.getInt("checksum"));
            }

            entry.getValue().close();

            return payment;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }


    @Nullable
    private DebitCardPayment getDebitCard(long id, @NotNull User user) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tablePaymentDebit, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            DebitCardPayment payment = null;
            rs.beforeFirst();
            while(rs.next()) {
                payment = new DebitCardPayment(id, user, rs.getString("iban"), rs.getString("bic"), rs.getString("owner"));
            }
            entry.getValue().close();

            return payment;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Nullable
    private InvoicePayment getInvoice(long id, @NotNull User user) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tablePaymentInvoice, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            rs.beforeFirst();
            InvoicePayment payment = null;
            while(rs.next()) {
                if(rs.getLong("id") == id && rs.getLong("user") == user.getId()) {
                    payment = new InvoicePayment(id, user);
                }
            }
            entry.getValue().close();

            return payment;
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return null;
    }

    public int insert(@NotNull PayPalPayment payment) {
        int insertCount = 0;
        Connection connection = database.getConnection();
        assert connection != null;

        try {
            long id = insertType(payment);

            PreparedStatement statement = connection.prepareStatement("INSERT INTO payment_paypal (id, mail, auth) VALUES (?, ?, ?)");
            statement.setLong(1, id);
            statement.setString(2, payment.getEmail());
            statement.setString(3, payment.getAuthCode());
            insertCount += statement.executeUpdate();

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return insertCount;
    }

    public int insert(@NotNull CreditCardPayment payment) {
        int insertCount = 0;
        Connection connection = database.getConnection();
        assert connection != null;

        try {
            long id = insertType(payment);

            PreparedStatement statement = connection.prepareStatement("INSERT INTO payment_credit (id, number, owner, expiration, checksum) VALUES (?, ?, ?, ?, ?)");
            statement.setLong(1, id);
            statement.setInt(2, payment.getNumber());
            statement.setString(3, payment.getOwner());
            statement.setDate(4, payment.getExpiration());
            statement.setInt(5, payment.getChecksum());

            insertCount += statement.executeUpdate();

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return insertCount;
    }

    public int insert(@NotNull DebitCardPayment payment) {
        int insertCount = 0;
        Connection connection = database.getConnection();
        assert connection != null;

        try {
            long id = insertType(payment);

            PreparedStatement statement = connection.prepareStatement("INSERT INTO payment_debit (id, iban, bic, owner) VALUES (?, ?, ?, ?)");
            statement.setLong(1, id);
            statement.setString(2, payment.getIban());
            statement.setString(3, payment.getBic());
            statement.setString(4, payment.getOwner());

            insertCount += statement.executeUpdate();

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return insertCount;
    }

    public int insert(@NotNull InvoicePayment payment) {
        int insertCount = 0;
        Connection connection = database.getConnection();
        assert connection != null;

        try {
            long id = insertType(payment);

            PreparedStatement statement = connection.prepareStatement("INSERT INTO payment_invoice (id) VALUES (?)");
            statement.setLong(1, id);

            insertCount += statement.executeUpdate();

            connection.close();
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return insertCount;
    }

    @Nullable
    private long insertType(@NotNull Payment payment) {
        Connection connection = database.getConnection();
        assert connection != null;

        try {
            PreparedStatement statement = connection.prepareStatement("INSERT INTO user_payment (type, user) VALUES (?, ?)", Statement.RETURN_GENERATED_KEYS);
            statement.setLong(1, payment.getType().getId());
            statement.setLong(2, payment.getUser().getId());

            statement.executeUpdate();

            ResultSet rs = statement.getGeneratedKeys();
            Long id = null;

            while(rs.next()) {
                id = rs.getLong("id");
            }

            connection.close();

            return id;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return -1;
    }

    public int delete(@NotNull Payment payment) {
        int deleteCount = 0;

        deleteCount += database.delete(payment.getId(), database.tableUserPayment);
        String type = payment.getType().getName();
        switch (type) {
            case "CreditCard":
                deleteCount += database.delete(payment.getId(), database.tablePaymentCredit);
                break;
            case "DebitCard":
                deleteCount += database.delete(payment.getId(), database.tablePaymentDebit);
                break;
            case "Invoice":
                deleteCount += database.delete(payment.getId(), database.tablePaymentInvoice);
                break;
            case "PayPal":
                deleteCount += database.delete(payment.getId(), database.tablePaymentPayPal);
                break;
        }
        return deleteCount;
    }

    @Nullable
    private Map.Entry<User, PaymentType> getMetadata(long id) {
        Map.Entry<ResultSet, Connection> entry = database.getResultSetById(database.tableUserPayment, id);
        assert entry != null;

        try {
            ResultSet rs = entry.getKey();
            rs.beforeFirst();
            Map.Entry<User, PaymentType> metadata = null;
            while(rs.next()) {
                metadata = new AbstractMap.SimpleEntry<>(database.user().get(rs.getLong("user")), getType(rs.getLong("type")));
            }
            entry.getValue().close();

            return metadata;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }
}